/*
   This file contains quick and dirty implementation of VESA library 
   for DJGPP platform

   Original author of vbelib: Nick Kurshev <nickols_k@mail.ru>,
   Neomagic TV out support by Rudolf Marek <r.marek et sh.cvut.cz>
   
   You can redistribute this file under terms and conditions
   of GNU General Public licence v2.
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <dos.h>
#include <crt0.h>
#include <dpmi.h>
#include <go32.h>
#include <sys/farptr.h>
#include <sys/nearptr.h>

#include "vbe.h"

//#define HAVE_VERBOSE_VAR 1
#ifdef HAVE_VERBOSE_VAR
extern int verbose;
#endif

static struct VesaProtModeInterface vbe_pm_info;

static struct VesaModeInfoBlock curr_mode_info;

struct realVesaProtModeInterface
{
  unsigned short SetWindowCall;
  unsigned short SetDisplayStart;
  unsigned short SetPaletteData;
  unsigned short iopl_ports;
}__attribute__((packed));

struct realVesaProtModeInterface *rm_info;

int vbeVersion(void)
{
  return 0x0002; // 0.2
}

static void kd_set_mode(int mode)
{
  __dpmi_regs r;
  static int old_mode;
  r.x.ax = 0x4f03;
  __dpmi_int(0x10, &r);
  old_mode = r.x.bx;
  if (mode == -1) return;
  r.x.ax = 0x4f02;
  r.x.bx = mode & 0x01ff;
  __dpmi_int(0x10, &r);
}

int vbeInit( void )
{
  if ((_crt0_startup_flags & _CRT0_FLAG_NEARPTR) == 0)
  {
	if (!__djgpp_nearptr_enable())
				return VBE_NO_NEARPTR;
	
	_crt0_startup_flags |= _CRT0_FLAG_NEARPTR;
  }

  memset(&vbe_pm_info,0,sizeof(struct VesaProtModeInterface));
  rm_info=NULL;
  vbeGetProtModeInfo(&vbe_pm_info);
  return VBE_OK;
}

int vbeDestroy( void ) 
{
  if (rm_info)
    free(rm_info);
  return VBE_OK;
}

#define FarToVirt(ptr) ((__djgpp_conventional_base + \
  (((unsigned long)(ptr) & 0xFFFF0000) >> 12) + \
  ((unsigned long)(ptr) & 0xFFFF)))

#define FarToFlat(ptr) (( \
  (((unsigned long)(ptr) & 0xFFFF0000) >> 12) + \
  ((unsigned long)(ptr) & 0xFFFF)))

#define InV2b(ptr) ((FarToFlat(ptr) < (dosbuf + 254)) && \
  (FarToFlat(ptr) >= (dosbuf + 34)))

#define InV3b(ptr) ((FarToFlat(ptr) < (dosbuf + 512)) && \
  (FarToFlat(ptr) >= (dosbuf + 256)))

int vbeGetControllerInfo(struct VbeInfoBlock *data)
{
  __dpmi_regs r;
  long dosbuf;
  dosbuf = __tb & 0xfffff;
  dosmemput(data, sizeof(struct VbeInfoBlock), dosbuf);
  r.x.ax = 0x4f00;
  r.x.di = dosbuf & 0xf;
  r.x.es = (dosbuf>>4) & 0xffff;
  if(__dpmi_int(0x10, &r) != 0)
    return VBE_VM86_FAIL;
  if (r.x.ax != 0x004f)
    return r.x.ax;
  dosmemget(dosbuf, sizeof(struct VbeInfoBlock), data);

  if (InV3b(data->OemStringPtr))
    data->OemStringPtr = (char *)((char *)data + (FarToFlat(data->OemStringPtr) - dosbuf));
  else
    data->OemStringPtr = (char *)FarToVirt(data->OemStringPtr);

  if (InV2b(data->VideoModePtr))
    data->VideoModePtr = (unsigned short *)((char *)data + (FarToFlat(data->VideoModePtr) - dosbuf));
  else
    data->VideoModePtr = (unsigned short *)FarToVirt(data->VideoModePtr);
 
  if (((data->VESAVersion >> 8) & 0xff) < 2)
  { 
     data->OemSoftwareRev = 0;
     data->OemVendorNamePtr = data->OemProductNamePtr = data->OemProductRevPtr = "";
     return VBE_OK;
  }

  if (InV3b(data->OemVendorNamePtr))
    data->OemVendorNamePtr = (char *)((char *)data + (FarToFlat(data->OemVendorNamePtr) - dosbuf));
  else
    data->OemVendorNamePtr = (char *)FarToVirt(data->OemVendorNamePtr);

  if (InV3b(data->OemProductNamePtr))
    data->OemProductNamePtr = (char *)((char *)data + (FarToFlat(data->OemProductNamePtr) - dosbuf));
  else
    data->OemProductNamePtr = (char *)FarToVirt(data->OemProductNamePtr);

  if (InV3b(data->OemProductRevPtr))
    data->OemProductRevPtr = (char *)((char *)data + (FarToFlat(data->OemProductRevPtr) - dosbuf));
  else
    data->OemProductRevPtr = (char *)FarToVirt(data->OemProductRevPtr);

return VBE_OK;
}

int vbeGetModeInfo(unsigned mode,struct VesaModeInfoBlock *data)
{
  __dpmi_regs r;
  long dosbuf;
  int i;
  mode &= 0x01ff;
  dosbuf = __tb & 0xfffff;
  dosmemput(data, sizeof(struct VesaModeInfoBlock), dosbuf);
  r.x.ax = 0x4f01;
  r.x.cx = mode;
  r.x.di = dosbuf & 0x0f;
  r.x.es = (dosbuf >> 4) & 0xffff;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  dosmemget(dosbuf, sizeof(struct VesaModeInfoBlock), data);
  return VBE_OK;
}

int vbeSetTV(unsigned int vesa_mode,unsigned int TV_mode) {
#define NR_MODES 8
unsigned int mode_table[NR_MODES] =    
			{0x101,0x103,0x111,0x114,0x120,0x121,0x122,0x123};
unsigned int tv_table[][NR_MODES] = {
	        	{0x201,0x202,0x211,0x212,0x221,0x231,0x222,0x232},
	        	{0x200,0x203,0x210,0x213,0x220,0x230,0xFFFF,0xFFFF}};

/*

Alternate mode map. If modes like 320x240 and 400x300 does not work, but
640x480 and 800x600 work, then try to replace above two lines with this
lines and write email to me if it works.
r.marek et sh.cvut.cz

	        	{0x201,0x202,0x211,0x212,0x222,0x223,0x224,0x225},
	        	{0x200,0x203,0x210,0x213,0x220,0x221,0xFFFF,0xFFFF}};

*/				 
   __dpmi_regs r;
  int i;
  for(i=0;((mode_table[i]!=(vesa_mode&0x3FF))&&(i<NR_MODES));i++) ;
  if(i==NR_MODES) return 0;
#ifdef HAVE_VERBOSE_VAR
  if(verbose > 1) printf("vbelib: Trying to set TV mode %x\n",tv_table[TV_mode][i]);
#endif
  r.x.ax = 0x4f14;
  r.x.bx = 0x20;
  r.x.cx = tv_table[TV_mode][i];
  r.x.dx = 0;
  r.x.di = 0;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  return r.x.ax;
}

int vbeSetMode(unsigned mode,struct VesaCRTCInfoBlock *data)
{
  __dpmi_regs r;
  long dosbuf;
  int i;
  if(data)
  {
    dosbuf = __tb & 0xfffff;
	dosmemput(data, sizeof(struct VesaCRTCInfoBlock), dosbuf);
	r.x.di = dosbuf & 0x0f;
	r.x.es = (dosbuf>>4) & 0xffff;
  }
  else
  {
	r.x.di = 0;
	r.x.es = 0;
  }
  r.x.ax = 0x4f02;
  r.x.bx = mode;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  /* For internal use */
  if (mode >= 0x0100)
  {
    if (vbeGetModeInfo(mode,&curr_mode_info) != VBE_OK)
		return VBE_BROKEN_BIOS;
  }
  return VBE_OK;
}

int vbeGetMode(unsigned *mode)
{
   __dpmi_regs r;
  r.x.ax = 0x4f03;
  __dpmi_int(0x10, &r);
  if(r.x.ax != 0x004f)
    return r.x.ax;
  *mode = r.d.ebx & 0xffff;
  return VBE_OK;
}

int vbeGetPixelClock(unsigned *mode,unsigned *pixel_clock) // in Hz
{
  __dpmi_regs r;
  r.x.ax = 0x4f0b;
  r.x.bx = 0;
  r.x.cx = *pixel_clock;
  r.x.dx = *mode;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  *pixel_clock = r.x.cx;
  return VBE_OK;
}

static int stateblocksize;

int vbeSaveState(void **data)
{
  __dpmi_regs r; 
  long dosbuf;
  int i;
  void *ptr; char *chr;
  dosbuf = __tb & 0xfffff;
  r.x.ax = 0x4f04;
  r.x.cx = 0x0f;
  r.x.dx = 0x00;
  if(__dpmi_int(0x10, &r) != 0)
    return VBE_VM86_FAIL;
  stateblocksize = r.x.bx * 64;
  r.x.ax = 0x4f04;
  r.x.cx = 0x0f;
  r.x.dx = 0x01;
  r.x.bx = dosbuf & 0x0f;
  r.x.es = (dosbuf >> 4) & 0xffff;
  if(__dpmi_int(0x10, &r) != 0)
    return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;	
  if((ptr = malloc(stateblocksize)) == NULL)
    return VBE_OUT_OF_MEM;
  dosmemget(dosbuf, stateblocksize, ptr);
  *data = ptr;
  return VBE_OK;
}

int vbeRestoreState(void *data)
{
  __dpmi_regs r;
  long dosbuf;
  int i;
  dosbuf = __tb & 0xfffff;
  dosmemput(data, stateblocksize, dosbuf);
  free(data);
  r.x.ax = 0x4f04;
  r.x.cx = 0x0f;
  r.x.dx = 0x02;
  r.x.bx = dosbuf & 0x0f;
  r.x.es = (dosbuf>>4) & 0xffff;
  if(__dpmi_int(0x10, &r) != 0)  
	return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  return VBE_OK;
}

int vbeGetWindow(unsigned *win_num)
{
  __dpmi_regs r;
  r.x.ax = 0x4f05;
  r.x.bx = (*win_num & 0x0f) | 0x0100;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  *win_num = r.x.dx;
  return VBE_OK;
}

int vbeSetWindow(unsigned win_num,unsigned win_gran)
{
  if(vbe_pm_info.SetWindowCall)
  {
    /* Don't verbose this stuff from performance reasons */
    /* 32-bit function call is much better of int 10h */
    __asm __volatile(
      "pushl  %%ebx\n"
      "movl   %0, %%ebx\n"
      "call   *%1\n"
      "popl   %%ebx"
      ::"S"(win_num & 0x0f),"D"(vbe_pm_info.SetWindowCall),
      "d"(win_gran):"memory");
    return VBE_OK;
  }
  else
  { 
    __dpmi_regs r;
    r.x.ax = 0x4f05;
    r.x.bx = win_num & 0x0f;
    r.x.dx = win_gran;
    if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
    if(r.x.ax != 0x004f)
      return r.x.ax;
    return VBE_OK;
  }
}

int vbeGetScanLineLength(unsigned *num_pixels,unsigned *num_bytes)
{
  __dpmi_regs r;
  r.x.ax = 0x4f06;
  r.x.bx = 1;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  if(num_bytes)  *num_bytes = r.x.bx;
  if(num_pixels) *num_pixels= r.x.cx;
  return VBE_OK;
}

int vbeGetMaxScanLines(unsigned *num_pixels,unsigned *num_bytes, unsigned *num_lines)
{
  __dpmi_regs r;
  r.x.ax = 0x4f06;
  r.x.bx = 3;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  if(num_bytes)  *num_bytes = r.x.bx;
  if(num_pixels) *num_pixels= r.x.cx;
  if(num_lines)  *num_lines = r.x.cx;
  return VBE_OK;
}

int vbeSetScanLineLength(unsigned num_pixels)
{
  __dpmi_regs r;
  r.x.ax = 0x4f06;
  r.x.bx = 0;
  r.x.cx = num_pixels;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  return VBE_OK;
}

int vbeSetScanLineLengthB(unsigned num_bytes)
{
  __dpmi_regs r;
  r.x.ax = 0x4f06;
  r.x.bx = 2;
  r.x.cx = num_bytes;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  return VBE_OK;
}

int vbeGetDisplayStart(unsigned *pixel_num,unsigned *scan_line)
{
  __dpmi_regs r;
  r.x.ax = 0x4f07;
  r.x.bx = 1;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  if(pixel_num) *pixel_num = r.x.cx;
  if(scan_line) *scan_line = r.x.dx;
  return VBE_OK;
}

int vbeSetDisplayStart(unsigned long offset, int vsync)
{
  __dpmi_regs r;
  if(vbe_pm_info.SetDisplayStart)
  {
    __asm __volatile(
      "pushl  %%ebx\n"
      "movl   %0, %%ebx\n"
      "call   *%1\n"
      "popl   %%ebx\n"
      ::"S"(0),"D"(vbe_pm_info.SetDisplayStart),
      "c"((offset>>2) & 0xffff),"d"((offset>>18)&0xffff):"memory");
  }
  else
  {
    r.x.ax = 0x4f07;
    r.x.bx = 0;
    r.x.cx = offset % curr_mode_info.BytesPerScanLine;
    r.x.dx = offset / curr_mode_info.BytesPerScanLine;
    if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
    if(r.x.ax != 0x004f)
      return r.x.ax;
  }
  if (vsync) while (!(inportb(0x3da) & 8));
  return VBE_OK;
}

int vbeSetScheduledDisplayStart(unsigned long offset, int vsync)
{
  if(vbe_pm_info.SetDisplayStart)
  {
    __asm __volatile(
      "pushl  %%ebx\n"
      "movl   %0, %%ebx\n"
      "call   *%1\n"
      "popl   %%ebx\n"
      ::"S"(0x02),"D"(vbe_pm_info.SetDisplayStart),
      "c"((offset>>2) & 0xffff),"d"((offset>>18)&0xffff):"memory");
  }
  else
  {
    __dpmi_regs r;
    r.x.ax = 0x4f07;
    r.x.bx = 0x02;
    r.d.ecx = offset;
    if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
    if(r.x.ax != 0x004f)
      return r.x.ax;
  }
  if (vsync) while (!(inportb(0x3da) & 8));
  return VBE_OK;
}

int vbeGetProtModeInfo(struct VesaProtModeInterface *pm_info)
{
  __dpmi_regs r;
  r.x.ax = 0x4f0a;
  r.x.bx = 0;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
  {
#ifdef HAVE_VERBOSE_VAR  
    if(verbose > 1) printf("vbelib: pm interface is not supported\n");
#endif
    return r.x.ax;
  }
  if((rm_info = malloc(r.x.cx)) == NULL)
  {
#ifdef HAVE_VERBOSE_VAR
    if(verbose > 1) printf("vbelib: malloc failed\n");
#endif
    return VBE_OUT_OF_MEM;
  }
  dosmemget (r.x.es * 16 + r.x.di, r.x.cx, rm_info);
  if(rm_info->iopl_ports)
  {
    unsigned short *io_ptr = (unsigned short *)((char *)rm_info + rm_info->iopl_ports);
    while (*io_ptr != 0xffff) io_ptr++; 
    io_ptr++;
    if (*io_ptr != 0xffff)
    {
      free(rm_info);
      rm_info = NULL;
#ifdef HAVE_VERBOSE_VAR  
      if(verbose > 1) printf("vbelib: mmio is not supported, pm interface is disabled\n");
#endif
      return VBE_VM86_FAIL;
    }
  }
  if((rm_info->SetWindowCall >= sizeof(struct realVesaProtModeInterface)) && (rm_info->SetWindowCall < r.x.cx - 1))
    pm_info->SetWindowCall = (void *)((char *)rm_info + rm_info->SetWindowCall);
  else
  {
    pm_info->SetWindowCall = NULL;
#ifdef HAVE_VERBOSE_VAR  
    if(verbose > 1) printf("vbelib: pm SetWindowCall is disabled\n");
#endif
  }
  if((rm_info->SetDisplayStart >= sizeof(struct realVesaProtModeInterface)) && (rm_info->SetDisplayStart < r.x.cx - 1))
    pm_info->SetDisplayStart = (void *)((char *)rm_info + rm_info->SetDisplayStart);
  else
  {
    pm_info->SetDisplayStart = NULL;
#ifdef HAVE_VERBOSE_VAR  
    if(verbose > 1) printf("vbelib: pm SetDisplayStart is disabled\n");
#endif
  }
  if((rm_info->SetPaletteData >= sizeof(struct realVesaProtModeInterface)) && (rm_info->SetPaletteData < r.x.cx - 1))
    pm_info->SetPaletteData  = (void *)((char *)rm_info + rm_info->SetPaletteData);
  else
  {
    pm_info->SetPaletteData = NULL;
#ifdef HAVE_VERBOSE_VAR  
    if(verbose > 1) printf("vbelib: pm SetPaletteData is disabled\n");
#endif
  }
#ifdef HAVE_VERBOSE_VAR  
  if(verbose > 1) printf("vbelib: pm interface is enabled\n");
#endif
  return VBE_OK;
}

int vbeWriteString(int x, int y, int attr, char *str)
{
  __dpmi_regs r;
  long dosbuf;
  int i;
  dosbuf = __tb & 0xfffff;
  dosmemput(str, strlen(str), dosbuf);
  r.x.bp = dosbuf & 0xf;
  r.x.es = (dosbuf>>4) & 0xffff;
  r.x.ax = 0x1300;
  r.x.cx = strlen(str);
  r.x.dx = ((y<<8)&0xff00)|(x&0xff);
  r.x.bx = attr;
  if(__dpmi_int(0x10, &r) != 0) return VBE_VM86_FAIL;
  if(r.x.ax != 0x004f)
    return r.x.ax;
  return VBE_OK;
}

void * vbeMapVideoBuffer(unsigned long phys_addr,unsigned long size)
{
  
  __dpmi_meminfo meminfo;
  
  meminfo.address = phys_addr; 
  meminfo.size = size;
#ifdef HAVE_VERBOSE_VAR
  if(verbose > 1) printf("vbelib: vbeMapVideoBuffer(%08lX,%08lX)\n",phys_addr,size);
#endif
  if(__dpmi_physical_address_mapping(&meminfo) != 0)
		return NULL;
	__dpmi_lock_linear_region(&meminfo);
  return (void *)((unsigned long)meminfo.address + __djgpp_conventional_base);
}

void vbeUnmapVideoBuffer(unsigned long linear_addr,unsigned long size)
{
  __dpmi_meminfo meminfo;
  
  meminfo.address = linear_addr;
  meminfo.size = size;
#ifdef HAVE_VERBOSE_VAR
  if(verbose > 1) printf("vbelib: vbeUnmapVideoBuffer(%08lX,%08lX)\n",linear_addr,size);
#endif
  if((meminfo.address - __djgpp_conventional_base) > 0x100000) 
  {
	__dpmi_unlock_linear_region(&meminfo);
	__dpmi_free_physical_address_mapping(&meminfo);
  }
}
